#!/usr/bin/env python

"""Command line FetchCommunicationService server

fetch-server.py is intended to make it as easy as possible to stand up
a FetchCommunicationService that serves Communication files stored on
the local filesystem.  When you run the command:

  fetch-server.py communications_source

this script should "just work", doing what a naive user would
reasonably expect should happen, regardless of whether
communications_source is:

- a directory of files (possibly with sub-directories, and possibly
   with files that aren't Communications)
- a TGZ file of Communications
- a ZIP file of Communications

"""
from __future__ import unicode_literals

import argparse
import logging
import os
import os.path
import zipfile

import humanfriendly

import concrete.version
from concrete.util.access import CommunicationContainerFetchHandler
from concrete.util.access_wrapper import FetchCommunicationServiceWrapper
from concrete.util.comm_container import (
    DirectoryBackedCommunicationContainer,
    MemoryBackedCommunicationContainer,
    ZipFileBackedCommunicationContainer)
from concrete.util import set_stdout_encoding


def main():
    set_stdout_encoding()

    parser = argparse.ArgumentParser(
        description="Command line FetchCommunicatonService server",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    parser.add_argument("communications_source",
                        help="A path to {1} a (possibly nested) directory of "
                        "Communications named using the convention "
                        "'[COMMUNICATION_ID].[comm|concrete|gz], "
                        "{2} a ZIP file of Communications, or "
                        "{3} a file (which can be a .tgz or .tar file) containing "
                        "one or more Communications, all of which will be read "
                        "into memory on startup")
    parser.add_argument("--host", default=None,
                        help="Network interface for server to listen on "
                        "(e.g. 'localhost', '0.0.0.0')")
    parser.add_argument("-p", "--port", type=int, default=9090,
                        help="Port for server to listen on")
    parser.add_argument("--max-file-size", type=str, default="1GiB",
                        help="Maximum size of (non-ZIP) files that can be read into memory "
                        "(e.g. '2G', '300MB')")
    parser.add_argument('-l', '--loglevel', '--log-level',
                        help='Logging verbosity level threshold (to stderr)',
                        default='info')
    concrete.version.add_argparse_argument(parser)
    args = parser.parse_args()

    logging.basicConfig(format='%(asctime)-15s %(levelname)s: %(message)s',
                        level=args.loglevel.upper())

    if os.path.isdir(args.communications_source):
        comm_container = DirectoryBackedCommunicationContainer(args.communications_source)
    elif zipfile.is_zipfile(args.communications_source):
        comm_container = ZipFileBackedCommunicationContainer(args.communications_source)
    else:
        max_file_size = humanfriendly.parse_size(args.max_file_size, binary=True)
        comm_container = MemoryBackedCommunicationContainer(args.communications_source,
                                                            max_file_size=max_file_size)
    logging.info('Using Communication Container of type %s' % type(comm_container))
    handler = CommunicationContainerFetchHandler(comm_container)

    fetch_service = FetchCommunicationServiceWrapper(handler)
    logging.info("Waiting for connections on port %d..." % args.port)
    fetch_service.serve(args.host, args.port)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals
import argparse
from time import time

from concrete.annotate import AnnotateCommunicationService
from concrete.metadata.ttypes import AnnotationMetadata
from concrete.util.annotate_wrapper import AnnotateCommunicationServiceWrapper
from concrete.util.net import find_port


class NoopAnnotateCommunicationService(AnnotateCommunicationService.Iface):
    METADATA_TOOL = 'No-op AnnotateCommunicationService'

    def annotate(self, communication):
        return communication

    def getMetadata(self,):
        metadata = AnnotationMetadata(tool=self.METADATA_TOOL,
                                      timestamp=int(time()))
        return metadata

    def getDocumentation(self):
        return '''\
        AnnotateCommunicationService that returns communication unmodified
        '''

    def shutdown(self):
        pass


def main():
    parser = argparse.ArgumentParser(description='Annotation service for '
                                                 'testing')

    parser.add_argument('--host', dest='host', action='store', type=str,
                        default='localhost', help='Host on which to listen')
    parser.add_argument('-p', '--port', dest='port', action='store', type=int,
                        default=None,
                        help='Port on which to listen (default: auto)')
    args = parser.parse_args()

    host = args.host
    if args.port is None:
        port = find_port()
    else:
        port = args.port

    print('Serving on %s:%d...' % (host, port))

    handler = NoopAnnotateCommunicationService()

    server = AnnotateCommunicationServiceWrapper(handler)
    server.serve(host, port)


if __name__ == '__main__':
    main()



from ..summarization import (
    SummarizationService)
from .service_wrapper import (
    ConcreteServiceClientWrapper,
    ConcreteServiceWrapper,
    HTTPConcreteServiceClientWrapper,
    SubprocessConcreteServiceWrapper)


SummarizationClientWrapper = type(
    'SummarizationClientWrapper',
    (ConcreteServiceClientWrapper,),
    {'concrete_service_class': SummarizationService})


HTTPSummarizationClientWrapper = type(
    'HTTPSummarizationClientWrapper',
    (HTTPConcreteServiceClientWrapper,),
    {'concrete_service_class': SummarizationService})


SummarizationServiceWrapper = type(
    'SummarizationServiceWrapper',
    (ConcreteServiceWrapper,),
    {'concrete_service_class': SummarizationService})


SubprocessSummarizationServiceWrapper = type(
    'SubprocessSummarizationServiceWrapper',
    (SubprocessConcreteServiceWrapper,),
    {'concrete_service_wrapper_class': SummarizationServiceWrapper})



"""
Helper functions for generating Concrete :class:`.UUID` objects
"""

# Force 'import uuid' to import the Python standard library module
# named "uuid", and not the "concrete.uuid" module
from __future__ import absolute_import
from __future__ import unicode_literals

import uuid as python_uuid

from thrift.Thrift import TType
from ..uuid.ttypes import UUID
from ..metadata.ttypes import AnnotationMetadata
from .mem_io import communication_deep_copy

from inspect import isroutine
import random
import logging

try:
    unicode
except NameError:
    unicode = str


def generate_UUID():
    """Return a Concrete UUID object with a random UUID4 value.

    Returns:
        a Concrete :class:`.UUID` object
    """
    return UUID(uuidString=str(python_uuid.uuid4()))


def hex_to_bin(h):
    '''
    Return binary encoding of hexadecimal string

    Args:
        h (str): string of hexadecimal characters

    Returns:
        an integer whose bit representation corresponds to the
        hexadecimal representation in h
    '''
    return int(h, 16)


def bin_to_hex(b, n=None):
    '''
    Return hexadecimal representation of binary value

    Args:
        b (int): integer whose bit representation will be converted
        n (int): length of returned hexadecimal string (the string will
            be left-padded with 0s if it is originally shorter than n;
            an exception will be thrown if it is longer; the string
            will be returned as-is if n is None)

    Returns:
        a string of hexadecimal characters representing the bit
        sequence in b, padded to be n characters long if n is not
        None

    Raises:
        ValueError: if `n` is not None and the hexadecimal string
            representing b is longer than n
    '''
    h = hex(b)[2:]
    if h.endswith('L'):
        h = h[:-1]
    if n is None:
        n = len(h)
    elif len(h) > n:
        raise ValueError('hex string "%s" is longer than %d chars' % (h, n))
    return ('0' * (n - len(h))) + h


def split_uuid(u):
    '''
    Split UUID string into three hexadecimal strings of sizes 12, 8, and
    12, returning those three strings (with hyphens stripped) in a
    tuple.

    Args:
        u (str): UUID string

    Returns:
        a tuple of three hexadecimal strings of sizes 12, 8, and 12,
        corresponding to the first two segments, middle two segments,
        and last segment of the input UUID string (with all hyphens
        stripped)

    Raises:
        ValueError: if UUID string is malformatted
    '''
    p = u.split('-')

    valid_input = (
        len(p) == 5 and
        len(p[0]) == 8 and
        len(p[1]) == 4 and
        len(p[2]) == 4 and
        len(p[3]) == 4 and
        len(p[4]) == 12
    )
    if not valid_input:
        raise ValueError(
            'uuid "%s" is not of the form xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx'
            % u
        )

    xs = p[0] + p[1]
    ys = p[2] + p[3]
    zs = p[4]

    return (xs, ys, zs)


def join_uuid(xs, ys, zs):
    '''
    Given three hexadecimal strings of sizes 12, 8, and 12, join them
    into a UUID string (inserting hyphens appropriately) and return the
    result.

    Args:
        xs (str): 12 hexadecimal characters that will form first
            two segments of the UUID string (size 8 and size 4
            respectively)
        ys (str): 8 hexadecimal characters that will form the third
            and fourth segment of the UUID string (each of size 4)
        zs (str): 12 hexadecimal characters that will form the last
            segment of the UUID string (size 12)

    Returns:
        string of size 36 (12 + 8 + 12 = 32, plus four hyphens inserted
        appropriately) comprising UUID formed from xs, ys, and zs

    Raises:
        ValueError: if xs, ys, or zs have incorrect length
    '''
    valid_input = (
        len(xs) == 12 and
        len(ys) == 8 and
        len(zs) == 12
    )
    if not valid_input:
        raise ValueError('uuid pieces do not have lengths 12, 8, 12')

    u = '-'.join((xs[:8], xs[8:], ys[:4], ys[4:], zs))

    return u


def generate_hex_unif(n):
    '''
    Generate and return random string of n hexadecimal characters.

    Args:
        n (int): number of characters of string to return

    Returns:
        string of n i.i.d. uniform hexadecimal characters
    '''
    return ''.join(random.choice('abcdef0123456789') for i in range(n))


def generate_uuid_unif():
    '''
    Generate and return random UUID string whose characters are drawn
    uniformly from the hexadecimal alphabet.

    Returns:
        string of hexadecimal characters drawn uniformly at random
        (delimited into five UUID-like segments by hyphens)
    '''
    return join_uuid(generate_hex_unif(12),
                     generate_hex_unif(8),
                     generate_hex_unif(12))


class _AnalyticUUIDGenerator(object):
    """
    Compressible UUID generator for a given analytic in a given
    Communication.

    Given a starting UUID, a consecutive (incrementing) sequence of
    UUIDs is generated; incrementation affects only the last twelve
    characters of the UUID and wraps around (those twelve characters
    only) when the end is reached.  The first four segments of the
    generated UUID are constant.
    """

    def __init__(self, u):
        '''
        Args:
            u (str): UUID string containing base characters and initial
                offset of UUIDs to generate
        '''
        (self._xs, ys, zs) = split_uuid(u)
        self._ys = generate_hex_unif(len(ys))
        self._z = hex_to_bin(generate_hex_unif(len(zs)))
        self._z_len = len(zs)
        self._z_bound = 2**(4 * len(zs))
        self.n = 0

    def __iter__(self):
        return self

    def __next__(self):
        """
        Generate and return a new concrete :class:`.UUID` object.
        (StopIteration will never be raised.)

        Returns:
            next concrete :class:`.UUID` object in the sequence
        """
        self._z = (self._z + 1) % self._z_bound
        self.n += 1
        return UUID(uuidString=join_uuid(
            self._xs, self._ys, bin_to_hex(self._z, self._z_len)
        ))

    def next(self):
        """
        Generate and return a new concrete :class:`.UUID` object.
        (StopIteration will never be raised.)

        Returns:
            next concrete :class:`.UUID` object in the sequence
        """
        return self.__next__()


class AnalyticUUIDGeneratorFactory(object):
    """
    Primary interface to generation of compressible UUIDs.  Each
    compressible UUID takes the form

        xxxxxxxx-xxxx-yyyy-yyyy-zzzzzzzzzzzz

    where each instance of x, y, or z is a hexadecimal digit, the group
    of x digits is shared across all annotations in a Communication, the
    group of y digits is shared across all annotations generated by a
    given analytic (by convention, :class:`.AnnotationMetadata` tool) in
    a given Communication, and the group of z digits is unique to each
    annotation (generated by a given analytic).  Thus all UUIDs in a
    Communication share the same first twelve hex digits and some UUIDs
    in a Communication share the same middle eight hex digits.
    Additionally, while the x and y components are generated uniformly
    at random, the z component for each analytic in a Communication
    starts at a uniform-at-random twelve hex digits for the first
    annotation and increments by one for each annotation thereafter.
    Thus the UUIDs of a Communication likely have many substrings in
    common and are easily compressed.  For example, we might find the
    following seven UUIDs in a Communication, corresponding to seven
    annotations split across two analytics:

        1bccb123-be45-7288-028a-4fdf3181ab51
        1bccb123-be45-7288-028a-4fdf3181ab52
        1bccb123-be45-7288-028a-4fdf3181ab53
        1bccb123-be45-df12-9c04-198eaa130a4e
        1bccb123-be45-df12-9c04-198eaa130a4f
        1bccb123-be45-df12-9c04-198eaa130a50
        1bccb123-be45-df12-9c04-198eaa130a51

    One generator factory should be created per Communication, and a new
    generator should be created from that factory for each analytic
    processing the communication.  Often each program represents a
    single analytic, so common usage is::

        augf = AnalyticUUIDGeneratorFactory(comm)
        aug = augf.create()
        for <each annotation object created by this analytic>:
            annotation = next(aug)
            <add annotation to communication>

    or if you're creating a new Communication::

        augf = AnalyticUUIDGeneratorFactory()
        aug = augf.create()
        comm = <create communication>
        comm.uuid = next(aug)
        for <each annotation object created by this analytic>:
            annotation = next(aug)
            <add annotation to communication>

    where the annotation objects might be objects of type
    :class:`.Parse`, :class:`.DependencyParse`, :class:`.TokenTagging`,
    :class:`.CommunicationTagging`, etc.
    """

    def __init__(self, comm=None):
        if comm is None:
            self.comm_uuid = generate_uuid_unif()
        else:
            self.comm_uuid = comm.uuid.uuidString

    def create(self):
        """
        Returns:
            A UUID generator for a new analytic.
        """
        return _AnalyticUUIDGenerator(self.comm_uuid)


def _filtered_getmembers(obj):
    """
    Generate key-value pairs of object members that may contain UUIDs.
    Over-generate, but filter the output enough that concrete objects
    can be traversed recursively using this function without leading to
    stack overflows or infinite loops.
    """

    for k in dir(obj):
        if not (k[0] == '_' or k == 'thrift_spec' or k == 'read' or
                k == 'write' or k == 'validate'):
            v = getattr(obj, k)
            if not (isroutine(v) or
                    isinstance(v, int) or isinstance(v, float) or
                    isinstance(v, str) or isinstance(v, unicode)):
                yield (k, v)


_FILTERED_TTYPES = set((TType.STRUCT, TType.LIST, TType.MAP, TType.SET))


def _fast_filtered_getmembers(obj):
    """Fast thrift-specific implementation of filtered_getmembers."""

    if hasattr(obj, 'thrift_spec'):
        for s in obj.thrift_spec:
            if s is not None:
                t = s[1]
                if t in _FILTERED_TTYPES:
                    k = s[2]
                    yield (k, getattr(obj, k))


class UUIDClustering(object):
    """
    Representation of the UUID instance clusters in a concrete
    communication (each cluster represents the set of nested members of
    the communication that reference or are identified by a given UUID).
    """

    def __init__(self, comm):
        self._clusters = dict()  # map: UUID -> set of nested members
        self._search(comm)

    def hashable_clusters(self):
        """Hashable version of UUIDClustering.

        Two UUIDClusterings c1 and c2 are equivalent
        (the two underlying Communications' UUID structures are
        equivalent) if and only if:

            c1.hashable_clusters() == c2.hashable_clusters()

        Returns:
            The set of unlabeled UUID clusters in a unique
            and hashable format.
        """
        return set(tuple(sorted(c)) for c in self._clusters.values())

    def _search(self, obj, prefix=()):
        """
        Search obj for UUIDs, calling _add_uuid_field when UUIDs are
        found and calling _search on other object members.
        When _search calls itself, prefix is appended with the object
        member name, forming a uniquely identifiable tuple
        representation of the path from the root object to a nested
        object member.
        """

        if isinstance(obj, UUID):
            self._add_uuid_field(obj.uuidString, prefix)
        elif isinstance(obj, list):
            for (i, v) in enumerate(obj):
                self._search(v, prefix + (('list', i),))
        elif isinstance(obj, set):
            raise ValueError('UUIDClustering does not support sets')
        elif isinstance(obj, dict):
            for (k, v) in obj.items():
                self._search(v, prefix + (('dict', k),))
        else:
            for (k, v) in _filtered_getmembers(obj):
                self._search(v, prefix + (k,))

    def _add_uuid_field(self, u, f):
        """
        Add UUID field f (a unique, hashable representation of the path
        from the root communication to a nested UUID object) to the UUID
        cluster indexed by UUID string u.
        """
        if u in self._clusters:
            self._clusters[u].add(f)
        else:
            self._clusters[u] = set([f])


class UUIDCompressor(object):
    '''
    Interface to replacing a Communication's UUIDs with compressible
    UUIDs.
    '''

    def __init__(self, single_analytic=False):
        '''
        Args:
            single_analytic (bool): True to generate new UUIDs using
                a single analytic for all annotations, false to use
                the annotation metadata tool name as the analytic id
        '''
        self.single_analytic = single_analytic

    def compress(self, comm):
        """
        Return a copy of a communication whose UUIDs have been
        replaced by compressible UUIDs using
        :class:`.AnalyticUUIDGeneratorFactory`.
        When this method returns this object's public member variable
        `uuid_map` will contain a dictionary mapping the original
        UUIDs to the new UUIDs.

        Args:
            comm (Communication): communication to be copied
                (the UUIDs of the copy will be made compressible)

        Returns:
            Communication: Deep copy of `comm` with compressed UUIDs
        """

        cc = communication_deep_copy(comm)
        self.augf = AnalyticUUIDGeneratorFactory(cc)
        self.augs = dict()
        self.uuid_map = dict()

        self._compress_uuids(cc)
        self._compress_uuid_refs(cc)

        return cc

    def _compress_uuids(self, obj, name_is_uuid=False, tool=None):
        """Generate new UUIDs in "uuid" fields and save mapping"""

        tool = self._get_tool(obj, tool)

        if name_is_uuid:
            if isinstance(obj, UUID):
                obj.uuidString = self._gen_uuid(obj, tool)
            else:
                logging.warning('uuid not instance of UUID')

        if not isinstance(obj, UUID):  # we already took care of "uuid"
            self._apply(
                lambda elt, elt_name_is_uuid: self._compress_uuids(
                    elt, name_is_uuid=elt_name_is_uuid, tool=tool
                ),
                obj
            )

    def _compress_uuid_refs(self, obj, name_is_uuid=False, tool=None):
        """Update UUID references (not in "uuid" fields) using saved mapping"""

        tool = self._get_tool(obj, tool)

        if isinstance(obj, UUID):
            if not name_is_uuid:
                obj.uuidString = self.uuid_map[obj.uuidString]
        else:
            self._apply(
                lambda elt, elt_name_is_uuid: self._compress_uuid_refs(
                    elt, name_is_uuid=elt_name_is_uuid, tool=tool
                ),
                obj
            )

    def _get_tool(self, obj, tool=None):
        """
        Return tool for this object, given the parent tool;
        update self.augs
        """

        if hasattr(obj, 'metadata'):
            if isinstance(obj.metadata, AnnotationMetadata):
                tool = obj.metadata.tool
            else:
                logging.warning('metadata not instance of AnnotationMetadata')
        if self.single_analytic:
            tool = None
        if tool not in self.augs:
            self.augs[tool] = self.augf.create()
        return tool

    def _gen_uuid(self, old_uuid, tool):
        """
        Return a new UUID for the provided tool, using self.augs;
        update self.uuid_map
        """

        aug = self.augs[tool]
        new_uuid = next(aug)
        if old_uuid.uuidString in self.uuid_map:
            raise ValueError('encountered UUID %s twice, aborting' %
                             old_uuid.uuidString)
        self.uuid_map[old_uuid.uuidString] = new_uuid.uuidString
        return new_uuid.uuidString

    @classmethod
    def _apply(cls, f, x):
        """
        Apply f to the members of x if it is a basic container type,
        otherwise apply f to x directly.
        """

        if isinstance(x, list):
            for elt in x:
                f(elt, False)
        elif isinstance(x, set):
            for elt in x:
                f(elt, False)
        elif isinstance(x, dict):
            for elt in x.values():
                f(elt, False)
        else:
            for (k, v) in _fast_filtered_getmembers(x):
                f(v, k == 'uuid')


def compress_uuids(comm, verify=False, single_analytic=False):
    """Create a copy of :class:`.Communication` `comm` with UUIDs
    converted according to the compressible UUID scheme

    Args:
        comm (Communication):
        verify (bool): If True, use a heuristic to verify the
            UUID link structure is preserved in the new Communication
        single_analytic (bool): If True, use a single analytic prefix
            for all UUIDs in `comm`.

    Returns:
        A 2-tuple containing the new :class:`.Communication`
        (converted using the compressible UUID scheme) and the
        :class:`UUIDCompressor` object used to perform the conversion.

    Raises:
        ValueError: If `verify` is True and `comm` has references added,
            raise because verification would cause an infinite loop.
    """

    if verify and hasattr(comm, 'tokenizationForUUID'):
        raise ValueError('cannot verify communication with references')

    uc = UUIDCompressor(single_analytic=single_analytic)

    new_comm = uc.compress(comm)

    num_old_uuids = len(set(uc.uuid_map.keys()))
    num_new_uuids = len(set(uc.uuid_map.values()))

    if verify:
        c1 = UUIDClustering(comm).hashable_clusters()
        c2 = UUIDClustering(new_comm).hashable_clusters()

        # Verification is c1 == c2;
        # also check UUID map lengths are the same as a sanity-check
        if num_old_uuids == num_new_uuids and c1 == c2:
            logging.info('verified %s (%d uuid instances, %d uuids)'
                         % (comm.id, sum(len(c) for c in c1), len(c1)))
        else:
            logging.error('%s failed verification' % comm.id)
            logging.error('uuid counts: %d -> %d'
                          % (num_old_uuids, num_new_uuids))
            logging.error('verified number of uuids: %d -> %d'
                          % (len(c1), len(c2)))
            logging.error('verified number of uuid instances: %d -> %d'
                          % (sum(map(len, c1)), sum(map(len, c2))))
            raise Exception('%s failed verification' % comm.id)

    else:
        if num_old_uuids != num_new_uuids:
            logging.warning('uuid counts are not the same (%d -> %d)'
                            % (num_old_uuids, num_new_uuids))

    return (new_comm, uc)



# -*- coding: utf-8 -*-
#
# Autogenerated by Thrift Compiler (0.11.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:coding=utf-8
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys

from thrift.transport import TTransport
all_structs = []


class Sound(object):
    """
    A sound wave. A separate optional field is defined for each
    suppported format. Typically, a Sound object will only define
    a single field.

    Note: we may want to have separate fields for separate channels
    (left vs right), etc.

    Attributes:
     - wav
     - mp3
     - sph
     - path: An absolute path to a file on disk where the sound file can be
    found. It is assumed that this path will be accessable from any
    machine that the system is run on (i.e., it should be a shared
    disk, or possibly a mirrored directory).
    """


    def __init__(self, wav=None, mp3=None, sph=None, path=None,):
        self.wav = wav
        self.mp3 = mp3
        self.sph = sph
        self.path = path

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.wav = iprot.readBinary()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRING:
                    self.mp3 = iprot.readBinary()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.STRING:
                    self.sph = iprot.readBinary()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.STRING:
                    self.path = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('Sound')
        if self.wav is not None:
            oprot.writeFieldBegin('wav', TType.STRING, 1)
            oprot.writeBinary(self.wav)
            oprot.writeFieldEnd()
        if self.mp3 is not None:
            oprot.writeFieldBegin('mp3', TType.STRING, 2)
            oprot.writeBinary(self.mp3)
            oprot.writeFieldEnd()
        if self.sph is not None:
            oprot.writeFieldBegin('sph', TType.STRING, 3)
            oprot.writeBinary(self.sph)
            oprot.writeFieldEnd()
        if self.path is not None:
            oprot.writeFieldBegin('path', TType.STRING, 4)
            oprot.writeString(self.path.encode('utf-8') if sys.version_info[0] == 2 else self.path)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(Sound)
Sound.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'wav', 'BINARY', None, ),  # 1
    (2, TType.STRING, 'mp3', 'BINARY', None, ),  # 2
    (3, TType.STRING, 'sph', 'BINARY', None, ),  # 3
    (4, TType.STRING, 'path', 'UTF8', None, ),  # 4
)
fix_spec(all_structs)
del all_structs



# -*- coding: utf-8 -*-
#
# Autogenerated by Thrift Compiler (0.11.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:coding=utf-8
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys
import concrete.services.Service
import logging
from .ttypes import *
from thrift.Thrift import TProcessor
from thrift.transport import TTransport
all_structs = []


class Iface(concrete.services.Service.Iface):
    """
    Service to fetch particular communications.
    """
    def fetch(self, request):
        """
        Parameters:
         - request
        """
        pass

    def getCommunicationIDs(self, offset, count):
        """
        Get a list of 'count' Communication IDs starting at 'offset'.  Implementations
        that do not provide this should throw an exception.

        Parameters:
         - offset
         - count
        """
        pass

    def getCommunicationCount(self):
        """
        Get the number of Communications this service searches over.  Implementations
        that do not provide this should throw an exception.
        """
        pass


class Client(concrete.services.Service.Client, Iface):
    """
    Service to fetch particular communications.
    """
    def __init__(self, iprot, oprot=None):
        concrete.services.Service.Client.__init__(self, iprot, oprot)

    def fetch(self, request):
        """
        Parameters:
         - request
        """
        self.send_fetch(request)
        return self.recv_fetch()

    def send_fetch(self, request):
        self._oprot.writeMessageBegin('fetch', TMessageType.CALL, self._seqid)
        args = fetch_args()
        args.request = request
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_fetch(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = fetch_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        if result.ex is not None:
            raise result.ex
        raise TApplicationException(TApplicationException.MISSING_RESULT, "fetch failed: unknown result")

    def getCommunicationIDs(self, offset, count):
        """
        Get a list of 'count' Communication IDs starting at 'offset'.  Implementations
        that do not provide this should throw an exception.

        Parameters:
         - offset
         - count
        """
        self.send_getCommunicationIDs(offset, count)
        return self.recv_getCommunicationIDs()

    def send_getCommunicationIDs(self, offset, count):
        self._oprot.writeMessageBegin('getCommunicationIDs', TMessageType.CALL, self._seqid)
        args = getCommunicationIDs_args()
        args.offset = offset
        args.count = count
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_getCommunicationIDs(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = getCommunicationIDs_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        if result.ex is not None:
            raise result.ex
        raise TApplicationException(TApplicationException.MISSING_RESULT, "getCommunicationIDs failed: unknown result")

    def getCommunicationCount(self):
        """
        Get the number of Communications this service searches over.  Implementations
        that do not provide this should throw an exception.
        """
        self.send_getCommunicationCount()
        return self.recv_getCommunicationCount()

    def send_getCommunicationCount(self):
        self._oprot.writeMessageBegin('getCommunicationCount', TMessageType.CALL, self._seqid)
        args = getCommunicationCount_args()
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_getCommunicationCount(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = getCommunicationCount_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        if result.ex is not None:
            raise result.ex
        raise TApplicationException(TApplicationException.MISSING_RESULT, "getCommunicationCount failed: unknown result")


class Processor(concrete.services.Service.Processor, Iface, TProcessor):
    def __init__(self, handler):
        concrete.services.Service.Processor.__init__(self, handler)
        self._processMap["fetch"] = Processor.process_fetch
        self._processMap["getCommunicationIDs"] = Processor.process_getCommunicationIDs
        self._processMap["getCommunicationCount"] = Processor.process_getCommunicationCount

    def process(self, iprot, oprot):
        (name, type, seqid) = iprot.readMessageBegin()
        if name not in self._processMap:
            iprot.skip(TType.STRUCT)
            iprot.readMessageEnd()
            x = TApplicationException(TApplicationException.UNKNOWN_METHOD, 'Unknown function %s' % (name))
            oprot.writeMessageBegin(name, TMessageType.EXCEPTION, seqid)
            x.write(oprot)
            oprot.writeMessageEnd()
            oprot.trans.flush()
            return
        else:
            self._processMap[name](self, seqid, iprot, oprot)
        return True

    def process_fetch(self, seqid, iprot, oprot):
        args = fetch_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = fetch_result()
        try:
            result.success = self._handler.fetch(args.request)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except concrete.services.ttypes.ServicesException as ex:
            msg_type = TMessageType.REPLY
            result.ex = ex
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("fetch", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_getCommunicationIDs(self, seqid, iprot, oprot):
        args = getCommunicationIDs_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = getCommunicationIDs_result()
        try:
            result.success = self._handler.getCommunicationIDs(args.offset, args.count)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except concrete.services.ttypes.NotImplementedException as ex:
            msg_type = TMessageType.REPLY
            result.ex = ex
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("getCommunicationIDs", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_getCommunicationCount(self, seqid, iprot, oprot):
        args = getCommunicationCount_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = getCommunicationCount_result()
        try:
            result.success = self._handler.getCommunicationCount()
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except concrete.services.ttypes.NotImplementedException as ex:
            msg_type = TMessageType.REPLY
            result.ex = ex
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("getCommunicationCount", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

# HELPER FUNCTIONS AND STRUCTURES


class fetch_args(object):
    """
    Attributes:
     - request
    """


    def __init__(self, request=None,):
        self.request = request

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.request = FetchRequest()
                    self.request.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('fetch_args')
        if self.request is not None:
            oprot.writeFieldBegin('request', TType.STRUCT, 1)
            self.request.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(fetch_args)
fetch_args.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'request', [FetchRequest, None], None, ),  # 1
)


class fetch_result(object):
    """
    Attributes:
     - success
     - ex
    """


    def __init__(self, success=None, ex=None,):
        self.success = success
        self.ex = ex

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.STRUCT:
                    self.success = FetchResult()
                    self.success.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 1:
                if ftype == TType.STRUCT:
                    self.ex = concrete.services.ttypes.ServicesException()
                    self.ex.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('fetch_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.STRUCT, 0)
            self.success.write(oprot)
            oprot.writeFieldEnd()
        if self.ex is not None:
            oprot.writeFieldBegin('ex', TType.STRUCT, 1)
            self.ex.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(fetch_result)
fetch_result.thrift_spec = (
    (0, TType.STRUCT, 'success', [FetchResult, None], None, ),  # 0
    (1, TType.STRUCT, 'ex', [concrete.services.ttypes.ServicesException, None], None, ),  # 1
)


class getCommunicationIDs_args(object):
    """
    Attributes:
     - offset
     - count
    """


    def __init__(self, offset=None, count=None,):
        self.offset = offset
        self.count = count

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I64:
                    self.offset = iprot.readI64()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.I64:
                    self.count = iprot.readI64()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('getCommunicationIDs_args')
        if self.offset is not None:
            oprot.writeFieldBegin('offset', TType.I64, 1)
            oprot.writeI64(self.offset)
            oprot.writeFieldEnd()
        if self.count is not None:
            oprot.writeFieldBegin('count', TType.I64, 2)
            oprot.writeI64(self.count)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(getCommunicationIDs_args)
getCommunicationIDs_args.thrift_spec = (
    None,  # 0
    (1, TType.I64, 'offset', None, None, ),  # 1
    (2, TType.I64, 'count', None, None, ),  # 2
)


class getCommunicationIDs_result(object):
    """
    Attributes:
     - success
     - ex
    """


    def __init__(self, success=None, ex=None,):
        self.success = success
        self.ex = ex

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.LIST:
                    self.success = []
                    (_etype17, _size14) = iprot.readListBegin()
                    for _i18 in range(_size14):
                        _elem19 = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                        self.success.append(_elem19)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 1:
                if ftype == TType.STRUCT:
                    self.ex = concrete.services.ttypes.NotImplementedException()
                    self.ex.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('getCommunicationIDs_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.LIST, 0)
            oprot.writeListBegin(TType.STRING, len(self.success))
            for iter20 in self.success:
                oprot.writeString(iter20.encode('utf-8') if sys.version_info[0] == 2 else iter20)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.ex is not None:
            oprot.writeFieldBegin('ex', TType.STRUCT, 1)
            self.ex.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(getCommunicationIDs_result)
getCommunicationIDs_result.thrift_spec = (
    (0, TType.LIST, 'success', (TType.STRING, 'UTF8', False), None, ),  # 0
    (1, TType.STRUCT, 'ex', [concrete.services.ttypes.NotImplementedException, None], None, ),  # 1
)


class getCommunicationCount_args(object):


    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('getCommunicationCount_args')
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(getCommunicationCount_args)
getCommunicationCount_args.thrift_spec = (
)


class getCommunicationCount_result(object):
    """
    Attributes:
     - success
     - ex
    """


    def __init__(self, success=None, ex=None,):
        self.success = success
        self.ex = ex

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.I64:
                    self.success = iprot.readI64()
                else:
                    iprot.skip(ftype)
            elif fid == 1:
                if ftype == TType.STRUCT:
                    self.ex = concrete.services.ttypes.NotImplementedException()
                    self.ex.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('getCommunicationCount_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.I64, 0)
            oprot.writeI64(self.success)
            oprot.writeFieldEnd()
        if self.ex is not None:
            oprot.writeFieldBegin('ex', TType.STRUCT, 1)
            self.ex.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(getCommunicationCount_result)
getCommunicationCount_result.thrift_spec = (
    (0, TType.I64, 'success', None, None, ),  # 0
    (1, TType.STRUCT, 'ex', [concrete.services.ttypes.NotImplementedException, None], None, ),  # 1
)
fix_spec(all_structs)
del all_structs




# -*- coding: utf-8 -*-
#
# Autogenerated by Thrift Compiler (0.11.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:coding=utf-8
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys
import concrete.uuid.ttypes

from thrift.transport import TTransport
all_structs = []


class AnnotationTaskType(object):
    """
    Annotation Tasks Types
    """
    TRANSLATION = 1
    NER = 2
    TOPICID = 3

    _VALUES_TO_NAMES = {
        1: "TRANSLATION",
        2: "NER",
        3: "TOPICID",
    }

    _NAMES_TO_VALUES = {
        "TRANSLATION": 1,
        "NER": 2,
        "TOPICID": 3,
    }


class AnnotationUnitType(object):
    """
    An annotation unit is the part of the communication to be annotated.
    """
    COMMUNICATION = 1
    SENTENCE = 2

    _VALUES_TO_NAMES = {
        1: "COMMUNICATION",
        2: "SENTENCE",
    }

    _NAMES_TO_VALUES = {
        "COMMUNICATION": 1,
        "SENTENCE": 2,
    }


class ServicesException(TException):
    """
    An exception to be used with Concrete services.

    Attributes:
     - message: The explanation (why the exception occurred)
     - serEx: The serialized exception
    """


    def __init__(self, message=None, serEx=None,):
        self.message = message
        self.serEx = serEx

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.message = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRING:
                    self.serEx = iprot.readBinary()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('ServicesException')
        if self.message is not None:
            oprot.writeFieldBegin('message', TType.STRING, 1)
            oprot.writeString(self.message.encode('utf-8') if sys.version_info[0] == 2 else self.message)
            oprot.writeFieldEnd()
        if self.serEx is not None:
            oprot.writeFieldBegin('serEx', TType.STRING, 2)
            oprot.writeBinary(self.serEx)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.message is None:
            raise TProtocolException(message='Required field message is unset!')
        return

    def __str__(self):
        return repr(self)

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class NotImplementedException(TException):
    """
    An exception to be used when an invoked method has
    not been implemented by the service.

    Attributes:
     - message: The explanation (why the exception occurred)
     - serEx: The serialized exception
    """


    def __init__(self, message=None, serEx=None,):
        self.message = message
        self.serEx = serEx

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.message = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRING:
                    self.serEx = iprot.readBinary()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('NotImplementedException')
        if self.message is not None:
            oprot.writeFieldBegin('message', TType.STRING, 1)
            oprot.writeString(self.message.encode('utf-8') if sys.version_info[0] == 2 else self.message)
            oprot.writeFieldEnd()
        if self.serEx is not None:
            oprot.writeFieldBegin('serEx', TType.STRING, 2)
            oprot.writeBinary(self.serEx)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.message is None:
            raise TProtocolException(message='Required field message is unset!')
        return

    def __str__(self):
        return repr(self)

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class AsyncContactInfo(object):
    """
    Contact information for the asynchronous communications.
    When a client contacts a server for a job that takes a significant amount of time,
    it is often best to implement this asynchronously.
    We do this by having the client stand up a server to accept the results and
    passing that information to the original server.
    The server may want to create a new thrift client on every request or maintain
    a pool of clients for reuse.

    Attributes:
     - host
     - port
    """


    def __init__(self, host=None, port=None,):
        self.host = host
        self.port = port

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.host = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.I32:
                    self.port = iprot.readI32()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('AsyncContactInfo')
        if self.host is not None:
            oprot.writeFieldBegin('host', TType.STRING, 1)
            oprot.writeString(self.host.encode('utf-8') if sys.version_info[0] == 2 else self.host)
            oprot.writeFieldEnd()
        if self.port is not None:
            oprot.writeFieldBegin('port', TType.I32, 2)
            oprot.writeI32(self.port)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.host is None:
            raise TProtocolException(message='Required field host is unset!')
        if self.port is None:
            raise TProtocolException(message='Required field port is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class AnnotationUnitIdentifier(object):
    """
    An annotation unit is the part of the communication to be annotated.
    It can be the entire communication or a particular sentence in the communication.
    If the sentenceID is null, the unit is the entire communication

    Attributes:
     - communicationId: Communication identifier for loading data
     - sentenceId: Sentence identifer if annotating sentences
    """


    def __init__(self, communicationId=None, sentenceId=None,):
        self.communicationId = communicationId
        self.sentenceId = sentenceId

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.communicationId = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRUCT:
                    self.sentenceId = concrete.uuid.ttypes.UUID()
                    self.sentenceId.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('AnnotationUnitIdentifier')
        if self.communicationId is not None:
            oprot.writeFieldBegin('communicationId', TType.STRING, 1)
            oprot.writeString(self.communicationId.encode('utf-8') if sys.version_info[0] == 2 else self.communicationId)
            oprot.writeFieldEnd()
        if self.sentenceId is not None:
            oprot.writeFieldBegin('sentenceId', TType.STRUCT, 2)
            self.sentenceId.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.communicationId is None:
            raise TProtocolException(message='Required field communicationId is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class ServiceInfo(object):
    """
    Each service is described by this info struct.
    It is for human consumption and for records of versions in deployments.

    Attributes:
     - name: Name of the service
     - version: Version string of the service.
    It is preferred that the services implement semantic versioning: http://semver.org/
    with version strings like x.y.z
     - description: Description of the service
    """


    def __init__(self, name=None, version=None, description=None,):
        self.name = name
        self.version = version
        self.description = description

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.name = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRING:
                    self.version = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.STRING:
                    self.description = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('ServiceInfo')
        if self.name is not None:
            oprot.writeFieldBegin('name', TType.STRING, 1)
            oprot.writeString(self.name.encode('utf-8') if sys.version_info[0] == 2 else self.name)
            oprot.writeFieldEnd()
        if self.version is not None:
            oprot.writeFieldBegin('version', TType.STRING, 2)
            oprot.writeString(self.version.encode('utf-8') if sys.version_info[0] == 2 else self.version)
            oprot.writeFieldEnd()
        if self.description is not None:
            oprot.writeFieldBegin('description', TType.STRING, 3)
            oprot.writeString(self.description.encode('utf-8') if sys.version_info[0] == 2 else self.description)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.name is None:
            raise TProtocolException(message='Required field name is unset!')
        if self.version is None:
            raise TProtocolException(message='Required field version is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(ServicesException)
ServicesException.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'message', 'UTF8', None, ),  # 1
    (2, TType.STRING, 'serEx', 'BINARY', None, ),  # 2
)
all_structs.append(NotImplementedException)
NotImplementedException.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'message', 'UTF8', None, ),  # 1
    (2, TType.STRING, 'serEx', 'BINARY', None, ),  # 2
)
all_structs.append(AsyncContactInfo)
AsyncContactInfo.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'host', 'UTF8', None, ),  # 1
    (2, TType.I32, 'port', None, None, ),  # 2
)
all_structs.append(AnnotationUnitIdentifier)
AnnotationUnitIdentifier.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'communicationId', 'UTF8', None, ),  # 1
    (2, TType.STRUCT, 'sentenceId', [concrete.uuid.ttypes.UUID, None], None, ),  # 2
)
all_structs.append(ServiceInfo)
ServiceInfo.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'name', 'UTF8', None, ),  # 1
    (2, TType.STRING, 'version', 'UTF8', None, ),  # 2
    (3, TType.STRING, 'description', 'UTF8', None, ),  # 3
)
fix_spec(all_structs)
del all_structs



"""Create a simple (valid) Communication suitable for testing purposes
"""
from __future__ import unicode_literals

import logging
import os
import re
import tempfile
import time

from ..communication.ttypes import Communication
from ..metadata.ttypes import AnnotationMetadata
from ..spans.ttypes import TextSpan
from ..structure.ttypes import (
    Section,
    Sentence,
    Token,
    Tokenization,
    TokenizationKind,
    TokenList
)
from .concrete_uuid import AnalyticUUIDGeneratorFactory
from .file_io import CommunicationWriter


AL_NONE = 'none'
AL_SECTION = 'section'
AL_SENTENCE = 'sentence'
AL_TOKEN = 'token'


def add_annotation_level_argparse_argument(parser):
    """Add an '--annotation-level' argument to an ArgumentParser

    The '--annotation-level argument specifies the level of
    concrete annotation to infer from whitespace in text.
    See :func:`create_comm` for details.

    Args:
        parser (argparse.ArgumentParser): the parser to add the argument
            to
    """
    parser.add_argument('--annotation-level', type=str,
                        choices=(AL_NONE, AL_SECTION, AL_SENTENCE, AL_TOKEN),
                        help='Level of concrete annotation to infer from'
                             ' whitespace in text (%s: no annotation,'
                             ' %s: sections inferred'
                             ' from double-newline, %s: "%s" and sentences'
                             ' inferred from single-newline, %s: "%s" and'
                             ' tokens inferred from remaining whitespace)' %
                             (AL_NONE,
                              AL_SECTION,
                              AL_SENTENCE, AL_SECTION,
                              AL_TOKEN, AL_SENTENCE))


def _split(s, delim):
    '''
    Split string and return list of tuples representing the pieces of
    the string and their offsets in the string.

    Args:
        s (str): string to split
        delim (str): delimiter by which to split string

    Returns:
        list of tuples representing the split pieces of the string;
        each tuple contains a string (a piece of `s`), an integer
        indicating where that string starts in `s`, and an integer
        indicating where that string ends in `s` (exclusive), in
        that order
    '''
    pieces = s.split(delim)
    indexed_pieces = []
    offset = 0
    for p in pieces:
        indexed_pieces.append((p, offset, offset + len(p)))
        offset += len(p) + len(delim)
    return indexed_pieces


def create_sentence(sen_text, sen_start, sen_end,
                    aug, metadata_tool, metadata_timestamp,
                    annotation_level):
    """Create :class:`.Sentence` from provided text and metadata.

    Lower-level routine (called indirectly by :func:`create_comm`)

    Args:
        sen_text (str): text to create sentence from
        sen_start (int): starting position of sentence in Communication
            text (inclusive)
        sen_end (int): ending position of sentence in Communication text
            (inclusive)
        aug (_AnalyticUUIDGenerator): compressible UUID generator for
            the analytic that generated this sentence
        metadata_tool (str): tool name of the analytic that generated
            this sentence
        metadata_timestamp (int): Time in seconds since the Epoch
        annotation_level (str): See :func:`create_comm` for details

    Returns:
        Concrete Sentence containing given text and metadata
    """

    sections = (annotation_level is not None) and (annotation_level != AL_NONE)
    sentences = sections and (annotation_level != AL_SECTION)
    tokens = sentences and (annotation_level != AL_SENTENCE)

    return Sentence(
        uuid=next(aug),
        textSpan=TextSpan(sen_start, sen_end),
        tokenization=Tokenization(
            uuid=next(aug),
            kind=TokenizationKind.TOKEN_LIST,
            metadata=AnnotationMetadata(
                tool=metadata_tool,
                timestamp=metadata_timestamp,
            ),
            tokenList=TokenList(tokenList=[
                Token(
                    tokenIndex=i,
                    text=match.group(),
                    textSpan=TextSpan(
                        start=match.start() + sen_start,
                        ending=match.end() + sen_start
                    ),
                )
                for (i, match)
                in enumerate([m for m in re.finditer(r'\S+', sen_text)])
            ]),
        ) if tokens else None,
    )


def create_section(sec_text, sec_start, sec_end, section_kind,
                   aug, metadata_tool, metadata_timestamp,
                   annotation_level):
    """Create :class:`.Section` from provided text and metadata.
    Section text will be split into sentence texts by newlines and
    each sentence will be created with a call to
    :func:`create_sentence`.

    Lower-level routine (called by :func:`create_comm`).

    Args:
        sec_text (str): text to create section from
        sec_start (int): starting position of section in Communication
            text (inclusive)
        sec_end (int): ending position of section in Communication text
            (inclusive)
        section_kind (str): value for `Section.kind` field to be set to
        aug (_AnalyticUUIDGenerator): compressible UUID generator for
            the analytic that generated this section
        metadata_tool (str): tool name of the analytic that generated
            this section
        metadata_timestamp (int): Time in seconds since the Epoch
        annotation_level (str): See :func:`create_comm` for details

    Returns:
        Concrete Section containing given text and metadata
    """

    sections = (annotation_level is not None) and (annotation_level != AL_NONE)
    sentences = sections and (annotation_level != AL_SECTION)

    return Section(
        uuid=next(aug),
        textSpan=TextSpan(sec_start, sec_end),
        kind=section_kind,
        sentenceList=(
            [
                create_sentence(sen_text,
                                sec_start + sen_start,
                                sec_start + sen_end,
                                aug, metadata_tool, metadata_timestamp,
                                annotation_level)
                for (sen_text, sen_start, sen_end) in _split(sec_text, '\n')
            ] if ('\n' in sec_text) or sec_text.strip() else []
        ) if sentences else None,
    )


def create_comm(comm_id, text='',
                comm_type='article', section_kind='passage',
                metadata_tool='concrete-python',
                metadata_timestamp=None,
                annotation_level=AL_TOKEN):
    """Create a simple, valid :class:`.Communication` from text.

    By default the text will be split by double-newlines into sections
    and then by single newlines into sentences within those sections.
    Each section will be created with a call to :func:`create_section`.

    `annotation_level` controls the amount of annotation that is added:

     - AL_NONE:      add no optional annotations (not even sections)
     - AL_SECTION:   add sections but not sentences
     - AL_SENTENCE:  add sentences but not tokens
     - AL_TOKEN:     add all annotations, up to tokens (the default)

    Args:
        comm_id (str): Communication id
        text (str): Communication text
        comm_type (str): Communication type
        section_kind (str): Section kind to set on all sections
        metadata_tool (str): tool name of analytic that generated
            this text
        metadata_timestamp (int): Time in seconds since the Epoch.
            If `None`, the current time will be used.
        annotation_level (str): string representing annotation
            level to add to communication (see above)

    Returns:
        Communication containing given text and metadata
    """

    if metadata_timestamp is None:
        metadata_timestamp = int(time.time())

    augf = AnalyticUUIDGeneratorFactory()
    aug = augf.create()

    sections = (annotation_level is not None) and (annotation_level != AL_NONE)

    return Communication(
        id=comm_id,
        uuid=next(aug),
        type=comm_type,
        text=text,
        metadata=AnnotationMetadata(
            tool=metadata_tool,
            timestamp=metadata_timestamp,
        ),
        sectionList=(
            [
                create_section(sec_text, sec_start, sec_end, section_kind,
                               aug, metadata_tool, metadata_timestamp,
                               annotation_level)
                for (sec_text, sec_start, sec_end) in _split_sections(text)
            ] if text.strip() else []
        ) if sections else None,
    )


def _split_sections(s):
    """Split string into sections when there are two or more "empty" lines

    Args:
        s (str): text to split

    Returns:
        list of tuples representing the split pieces of the section;
        each tuple contains a string (a piece of `s`), an integer
        indicating where that section starts in `s`, and an integer
        indicating where that section ends in `s` (exclusive), in
        that order
    """
    # Ignore whitespace at beginning/end of document
    m = re.match(r'^(?:\s*\r?\n)*(.*?)(?:\s*\r?\n)*$', s, re.DOTALL)
    stripped_start_offset = m.span(1)[0]
    stripped_end_offset = m.span(1)[1]
    stripped = s[stripped_start_offset:stripped_end_offset]

    offsets = []
    offsets.append(stripped_start_offset)
    for sec_break in [sm.span() for sm in re.finditer(r'(?:\s*\r?\n){2,}', stripped, re.DOTALL)]:
        offsets.append(sec_break[0] + stripped_start_offset)
        offsets.append(sec_break[1] + stripped_start_offset)
    offsets.append(stripped_end_offset)

    sections = []
    it = iter(offsets)
    for (start, end) in zip(it, it):
        sections.append((s[start:end], start, end))

    return sections


def create_simple_comm(comm_id, sentence_string="Super simple sentence ."):
    """Create a simple (valid) :class:`.Communication` suitable for testing purposes

    The Communication will have a single :class:`.Section` containing
    a single :class:`.Sentence`.

    Args:
        comm_id (str): Communication id
        sentence_string (str): Communication text

    Returns:
        Communication containing given text and having the given id
    """
    logging.warning('create_simple_comm will be removed in a future'
                    ' release, please use create_comm instead')

    toolname = "TEST"
    timestamp = int(time.time())

    augf = AnalyticUUIDGeneratorFactory()
    aug = augf.create()

    comm = Communication(
        id=comm_id,
        metadata=AnnotationMetadata(tool=toolname, timestamp=timestamp),
        type=toolname,
        uuid=next(aug)
    )

    tokenization = Tokenization(
        kind=TokenizationKind.TOKEN_LIST,
        metadata=AnnotationMetadata(tool=toolname, timestamp=timestamp),
        tokenList=TokenList(
            tokenList=[]),
        uuid=next(aug)
    )
    token_string_list = sentence_string.split()
    for i, token_string in enumerate(token_string_list):
        tokenization.tokenList.tokenList.append(Token(text=token_string,
                                                      tokenIndex=i))

    sentence = Sentence(
        textSpan=TextSpan(0, len(sentence_string)),
        tokenization=tokenization,
        uuid=next(aug)
    )

    section = Section(
        kind="SectionKind",
        sentenceList=[sentence],
        textSpan=TextSpan(0, len(sentence_string)),
        uuid=next(aug)
    )

    comm.sectionList = [section]
    comm.text = sentence_string

    return comm


class SimpleCommTempFile(object):
    """DEPRECATED. Please use :func:`create_comm` instead.

    Class representing a temporary file of sample concrete objects.
    Designed to facilitate testing.

    Attributes:
        path (str): path to file
        communications (Communication[]): List of communications that were written to file

    Usage::

        from concrete.util import CommunicationReader
        with SimpleCommTempFile(n=3, id_fmt='temp-%d') as f:
            reader = CommunicationReader(f.path)
            for (orig_comm, comm_path_pair) in zip(f.communications, reader):
                print(orig_comm.id)
                print(orig_comm.id == comm_path_pair[0].id)
                print(f.path == comm_path_pair[1])
    """

    def __init__(self, n=10, id_fmt='temp-%d',
                 sentence_fmt='Super simple sentence %d .',
                 writer_class=CommunicationWriter, suffix='.concrete'):
        """
        Create temp file and write communications.

        Args:
            n:i     number of communications to write
            id_fmt: format string used to generate communication IDs;
                    should contain one instance of %d, which will be
                    replaced by the number of the communication
            sentence_fmt: format string used to generate communication
                    IDs; should contain one instance of %d, which will
                    be replaced by the number of the communication
            writer_class: CommunicationWriter or CommunicationWriterTGZ
            suffix: file path suffix (you probably want to choose this
                    to match writer_class)
        """
        logging.warning('SimpleCommTempFile will be removed in a future'
                        ' release, please use create_comm instead')

        (fd, path) = tempfile.mkstemp(suffix=suffix)
        os.close(fd)
        self.path = path
        self.communications = []
        w = writer_class()
        w.open(path)
        for i in range(n):
            comm = create_simple_comm(id_fmt % i, sentence_fmt % i)
            self.communications.append(comm)
            w.write(comm)
        w.close()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if os.path.exists(self.path):
            os.remove(self.path)



from __future__ import unicode_literals

__version__ = '4.14.1'


def concrete_library_version():
    return __version__


def concrete_schema_version():
    return '.'.join(__version__.split('.')[0:2])


def add_argparse_argument(parser):
    from .util.thrift_factory import is_accelerated
    parser.add_argument(
        '--version', action='version',
        version='concrete-python %s (schema %s) (%s)' % (
            concrete_library_version(),
            concrete_schema_version(),
            'accelerated' if is_accelerated() else 'unaccelerated'
        )
    )



from __future__ import unicode_literals
from thrift.transport import TTransport
from thrift.transport import TSocket
from thrift.protocol import TCompactProtocol

from concrete.util import SubprocessSearchServiceWrapper
from concrete.util import find_port


from time import time

from concrete.search import SearchService
from concrete import (
    SearchType, SearchQuery, SearchResult, SearchResultItem
)
from concrete import UUID
from concrete import AnnotationMetadata


class FooSearch(SearchService.Iface):
    METADATA_TOOL = 'Foo Search'

    def search(self, search_query):
        return SearchResult(
            uuid=UUID(uuidString='12345678-1234-5678-1234-567812345678'),
            searchResultItems=[
                SearchResultItem(communicationId=term, score=42.)
                for term in search_query.terms
            ],
            metadata=AnnotationMetadata(tool=self.METADATA_TOOL,
                                        timestamp=int(time()))
        )

    def shutdown(self):
        pass


def test_search_communications():
    impl = FooSearch()
    host = 'localhost'
    port = find_port()
    timeout = 5

    terms = ['foo', 'bar']
    query = SearchQuery(type=SearchType.COMMUNICATIONS,
                        terms=[t for t in terms])

    with SubprocessSearchServiceWrapper(impl, host, port, timeout=timeout):
        transport = TSocket.TSocket(host, port)
        transport = TTransport.TFramedTransport(transport)
        protocol = TCompactProtocol.TCompactProtocolAccelerated(transport)

        cli = SearchService.Client(protocol)
        transport.open()
        res = cli.search(query)
        transport.close()

        assert res.uuid.uuidString == '12345678-1234-5678-1234-567812345678'
        assert len(res.searchResultItems) == 2
        assert res.searchResultItems[0].communicationId == 'foo'
        assert res.searchResultItems[0].score == 42.
        assert res.searchResultItems[1].communicationId == 'bar'
        assert res.searchResultItems[1].score == 42.
        assert res.metadata.tool == 'Foo Search'



"""Functions used by `concrete_inspect.py` to print data in a Communication.

The function implementations provide useful examples of how to
interact with many different Concrete datastructures.
"""
from __future__ import print_function
from __future__ import unicode_literals

import logging
from collections import defaultdict
from operator import attrgetter

from .util.metadata import filter_unnone, tool_to_filter
from .util.unnone import lun
from .util.tokenization import get_tokenizations, get_token_taggings

try:
    unicode
except NameError:
    unicode = str


def _get_tagged_token_strs_by_token_index(tagged_tokens, num_tokens):
    '''
    Return list of `num_tokens` strings:
    the specified token tag at each token index if a tag exists at that
    index, else the empty string.

    Args:
        tagged_tokens (list): list of :class:`.TaggedToken` objects
        num_tokens (int): number of tokens

    Returns:
        list of token tags (or empty strings for untagged tokens)
    '''
    tagged_tokens_by_token_index = dict(
        (tagged_token.tokenIndex, tagged_token)
        for tagged_token in tagged_tokens
    )
    return [
        (
            tagged_tokens_by_token_index[token_index].tag
            if token_index in tagged_tokens_by_token_index
            else u''
        )
        for token_index in range(num_tokens)
    ]


def print_conll_style_tags_for_communication(
        comm, char_offsets=False, dependency=False, lemmas=False, ner=False,
        pos=False, starts=False, endings=False,
        dependency_tool=None, dependency_parse_filter=None,
        lemmas_tool=None, lemmas_filter=None,
        ner_tool=None, ner_filter=None,
        pos_tool=None, pos_filter=None,
        other_tags=None):

    """
    Print 'CoNLL-style' tags for the tokens in a Communication.
    If column is requested (for example, `ner` is set to `True`) but
    there is no such annotation in the communication, that column is
    not printed (the header is not printed either).  If there is more
    than one such annotation in the communication, one column is printed
    for each annotation.  In the event of differing numbers of
    annotations per Tokenization, all annotations are printed, but it is
    not guaranteed that the columns of two different tokenizations
    correspond to one another.

    Args:
        comm (Communication):
        char_offsets (bool): Flag for printing token text specified by
          a :class:`.Token`'s (optional) :class:`.TextSpan`
        dependency (bool): Flag for printing dependency parse HEAD tags
        dependency_tool (str): Deprecated.
            If not `None`, only print information for
            :class:`.DependencyParse` objects if they have a matching
            `metadata.tool` field
        dependency_parse_filter (func): If not None, print information
            for only those :class:`.DependencyParse` objects that pass
            this filter.  Should be a function that takes a list of
            annotations (objects with metadata fields) and returns a
            list of annotations (possibly filtered and re-ordered).
        lemmas (bool): Flag for printing lemma tags
            (:class:`.TokenTagging` objects of type LEMMA)
        lemmas_tool (str): Deprecated.
            If not `None`, only print information for
            :class:`.TokenTagging` objects of type LEMMA if they have
            a matching `metadata.tool` field
        lemmas_filter (func): If not None, print information
            for only those LEMMA taggings that pass
            this filter.  Should be a function that takes a list of
            annotations (objects with metadata fields) and returns a
            list of annotations (possibly filtered and re-ordered).
        ner (bool): Flag for printing Named Entity Recognition tags
            (:class:`.TokenTagging` objects of type NER)
        ner_tool (str): Deprecated.
            If not `None`, only print information for
            :class:`.TokenTagging` objects of type NER if they have
            a matching `metadata.tool` field
        ner_filter (func): If not None, print information
            for only those NER taggings that pass
            this filter.  Should be a function that takes a list of
            annotations (objects with metadata fields) and returns a
            list of annotations (possibly filtered and re-ordered).
        pos (bool): Flag for printing Part-of-Speech tags
            (:class:`.TokenTagging` objects of type POS)
        pos_tool (str): Deprecated.
            If not `None`, only print information for
            :class:`.TokenTagging` objects of type POS if they have
            a matching `metadata.tool` field
        pos_filter (func): If not None, print information
            for only those POS taggings that pass
            this filter.  Should be a function that takes a list of
            annotations (objects with metadata fields) and returns a
            list of annotations (possibly filtered and re-ordered).
        other_tags (dict): Map of other tagging types to print (as keys)
            to annotation filters, or None.  If the value (annotation
            filter) of a given tagging type is not None, print
            information for only those taggings that pass
            the filter (should be a function that takes a list of
            annotations (objects with metadata fields) and returns a
            list of annotations (possibly filtered and re-ordered)).
    """
    dependency_parse_filter = filter_unnone(tool_to_filter(
        dependency_tool, dependency_parse_filter))
    lemmas_filter = filter_unnone(tool_to_filter(
        lemmas_tool, lemmas_filter))
    ner_filter = filter_unnone(tool_to_filter(
        ner_tool, ner_filter))
    pos_filter = filter_unnone(tool_to_filter(
        pos_tool, pos_filter))

    if other_tags is None:
        other_tags = dict()

    header_fields_by_tokenization = []
    field_lists_by_tokenization = []

    for tokenization in get_tokenizations(comm):
        header_fields = []
        field_lists = []

        header_fields.append(u'INDEX')
        field_lists.append([
            unicode(i + 1)
            for i in range(len(tokenization.tokenList.tokenList))
        ])

        header_fields.append(u'TOKEN')
        field_lists.append([
            token.text
            for token in tokenization.tokenList.tokenList
        ])

        if char_offsets:
            header_fields.append(u'CHAR')
            field_lists.append(
                _get_char_offset_tags_for_tokenization(comm, tokenization))

        if lemmas:
            for token_tagging in lemmas_filter(
                    get_token_taggings(tokenization, u'LEMMA')):
                header_fields.append(u'LEMMA')
                field_lists.append(
                    _get_tagged_token_strs_by_token_index(
                        token_tagging.taggedTokenList,
                        len(tokenization.tokenList.tokenList)))

        if pos:
            for token_tagging in pos_filter(
                    get_token_taggings(tokenization, u'POS')):
                header_fields.append(u'POS')
                field_lists.append(
                    _get_tagged_token_strs_by_token_index(
                        token_tagging.taggedTokenList,
                        len(tokenization.tokenList.tokenList)))

        if ner:
            for token_tagging in ner_filter(
                    get_token_taggings(tokenization, u'NER')):
                header_fields.append(u'NER')
                field_lists.append([
                    (tag if tag != u'NONE' else u'')
                    for tag in _get_tagged_token_strs_by_token_index(
                        token_tagging.taggedTokenList,
                        len(tokenization.tokenList.tokenList)
                    )
                ])

        if dependency:
            for conll_tag_pair_list in _get_conll_tags_for_tokenization(
                    tokenization,
                    dependency_parse_filter=dependency_parse_filter):
                header_fields.append(u'HEAD')
                field_lists.append([p[0] for p in conll_tag_pair_list])
                header_fields.append(u'DEPREL')
                field_lists.append([p[1] for p in conll_tag_pair_list])

        for (tag, tag_filter) in other_tags.items():
            _filter = filter_unnone(tag_filter)
            for token_tagging in _filter(get_token_taggings(tokenization, tag)):
                header_fields.append(tag)
                field_lists.append(
                    _get_tagged_token_strs_by_token_index(
                        token_tagging.taggedTokenList,
                        len(tokenization.tokenList.tokenList)))

        if starts:
            header_fields.append(u'START')
            field_lists.append(
                _get_start_tags_for_tokenization(comm, tokenization))

        if endings:
            header_fields.append(u'ENDING')
            field_lists.append(
                _get_ending_tags_for_tokenization(comm, tokenization))

        header_fields_by_tokenization.append(header_fields)
        field_lists_by_tokenization.append(field_lists)

    if len(set(map(tuple, header_fields_by_tokenization))) > 1:
        logging.warning(
            'communication does not have same taggings for each tokenization')

    def _max_num_header_fields(header_field):
        return max(
            header_fields.count(header_field)
            for header_fields
            in header_fields_by_tokenization
        )

    overall_header_fields = (
        ([u'INDEX'] * _max_num_header_fields(u'INDEX')) +
        ([u'TOKEN'] * _max_num_header_fields(u'TOKEN')) +
        ([u'CHAR'] * _max_num_header_fields(u'CHAR')) +
        ([u'LEMMA'] * _max_num_header_fields(u'LEMMA')) +
        ([u'POS'] * _max_num_header_fields(u'POS')) +
        ([u'NER'] * _max_num_header_fields(u'NER')) +
        ([u'HEAD', u'DEPREL'] * _max_num_header_fields(u'HEAD')) +
        ([u'START'] * _max_num_header_fields(u'START')) +
        ([u'ENDING'] * _max_num_header_fields(u'ENDING'))
    )
    for tag in other_tags:
        overall_header_fields.extend([tag] * _max_num_header_fields(tag))

    print(u'\t'.join(overall_header_fields))
    print(u'\t'.join(u'-' * len(header) for header in overall_header_fields))

    for (header_fields, field_lists) in zip(
            header_fields_by_tokenization, field_lists_by_tokenization):
        for row in zip(*field_lists):
            def _generate_row():
                overall_field_num = 0
                for (field_num, header) in enumerate(header_fields):
                    while header != overall_header_fields[overall_field_num]:
                        yield u''
                        overall_field_num += 1
                    yield row[field_num]
                    overall_field_num += 1
            print('\t'.join(_generate_row()))
        print()


def _print_entity_mention_content(em, prefix=''):
    '''
    Print information for :class:`.EntityMention` `em`, prefixing each
    line by `prefix`.

    Args:
        em (EntityMention):
        prefix (str):
    '''
    print(prefix + u"tokens:     %s" % (
        u" ".join(_get_tokens_for_entityMention(em))))
    if em.text:
        print(prefix + u"text:       %s" % em.text)
    print(prefix + u"entityType: %s" % em.entityType)
    print(prefix + u"phraseType: %s" % em.phraseType)


def print_entities(comm, tool=None, entity_set_filter=None):
    """Print information for :class:`.Entity` objects and their
    associated :class:`.EntityMention` objects

    Args:
        comm (Communication):
        tool (str): Deprecated.
                    If not `None`, only print information for
                    :class:`.EntitySet` objects with a matching
                    `metadata.tool` field
        entity_set_filter (func): If not None, print information
            for only those :class:`.EntitySet` objects that pass
            this filter.  Should be a function that takes a list of
            annotations (objects with metadata fields) and returns a
            list of annotations (possibly filtered and re-ordered).
    """
    _filter = filter_unnone(tool_to_filter(tool, entity_set_filter))
    for (entitySet_index, entitySet) in enumerate(lun(comm.entitySetList)):
        if _filter([entitySet]):
            print(u"Entity Set %d (%s):" % (entitySet_index,
                                            entitySet.metadata.tool))
            for entity_index, entity in enumerate(entitySet.entityList):
                print(u"  Entity %d-%d:" % (entitySet_index, entity_index))
                for em_index, em in enumerate(entity.mentionList):
                    print(u"      EntityMention %d-%d-%d:" % (
                        entitySet_index, entity_index, em_index))
                    _print_entity_mention_content(em, prefix=' ' * 10)
                    for (cm_index, cm) in enumerate(em.childMentionList):
                        print(u"          child EntityMention #%d:" % cm_index)
                        _print_entity_mention_content(cm, prefix=' ' * 14)
                print()
            print()


def print_metadata(comm, tool=None, annotation_filter=None):
    """Print metadata tools used to annotate Communication

    Args:
        comm (Communication):
        tool (str): Deprecated.
                    If not `None`, only print :class:`.AnnotationMetadata`
                    information for objects with a matching
                    `metadata.tool` field
        annotation_filter (func): If not None, print information
            for only those objects that pass this filter.  Should be a
            function that takes a list of annotations (objects with
            metadata fields) and returns a list of annotations (possibly
            filtered and re-ordered).
    """
    _filter = filter_unnone(tool_to_filter(tool, annotation_filter))
    if _filter([comm]):
        print(u"Communication:  %s\n" % comm.metadata.tool)

    dependency_parse_tools = set()
    parse_tools = set()
    tokenization_tools = set()
    token_tagging_tools = set()

    tokenizations = get_tokenizations(comm)
    tokenization_tools.update(
        ann.metadata.tool
        for ann in _filter(tokenizations))
    for tokenization in tokenizations:
        token_tagging_tools.update(
            ann.metadata.tool
            for ann in _filter(lun(tokenization.tokenTaggingList)))
        dependency_parse_tools.update(
            ann.metadata.tool
            for ann in _filter(lun(tokenization.dependencyParseList)))
        parse_tools.update(
            ann.metadata.tool
            for ann in _filter(lun(tokenization.parseList)))

    communication_tagging_tools = set(
        ann.metadata.tool
        for ann in _filter(lun(comm.communicationTaggingList)))

    if tokenization_tools:
        for toolname in sorted(tokenization_tools):
            print(u"  Tokenization:  %s" % toolname)
        print()
    if dependency_parse_tools:
        for toolname in sorted(dependency_parse_tools):
            print(u"    Dependency Parse:  %s" % toolname)
        print()
    if parse_tools:
        for toolname in sorted(parse_tools):
            print(u"    Parse:  %s" % toolname)
        print()
    if token_tagging_tools:
        for toolname in sorted(token_tagging_tools):
            print(u"    TokenTagging:  %s" % toolname)
        print()

    if comm.entityMentionSetList:
        for i, em_set in enumerate(comm.entityMentionSetList):
            if _filter([em_set]):
                print(u"  EntityMentionSet #%d:  %s" % (
                    i, em_set.metadata.tool))
        print()
    if comm.entitySetList:
        for i, entitySet in enumerate(comm.entitySetList):
            if _filter([entitySet]):
                print(u"  EntitySet #%d:  %s" % (
                    i, entitySet.metadata.tool))
        print()
    if comm.situationMentionSetList:
        for i, sm_set in enumerate(comm.situationMentionSetList):
            if _filter([sm_set]):
                print(u"  SituationMentionSet #%d:  %s" % (
                    i, sm_set.metadata.tool))
        print()
    if comm.situationSetList:
        for i, situationSet in enumerate(comm.situationSetList):
            if _filter([situationSet]):
                print(u"  SituationSet #%d:  %s" % (
                    i, situationSet.metadata.tool))
        print()

    if communication_tagging_tools:
        for toolname in sorted(communication_tagging_tools):
            print(u"  CommunicationTagging:  %s" % toolname)
        print()


def print_sections(comm, tool=None, communication_filter=None):
    """
    Print information for all :class:`.Section` object, according to
    their spans.

    Args:
        comm (Communication):
        tool (str): Deprecated.
                    If not `None`, only print information for
                    :class:`.Section` objects with a matching
                    `metadata.tool` field
        communication_filter (func): If not None, print information
            for only those :class:`.Communication` objects that pass
            this filter.  Should be a function that takes a list of
            annotations (objects with metadata fields) and returns a
            list of annotations (possibly filtered and re-ordered).
    """
    _filter = filter_unnone(tool_to_filter(tool, communication_filter))
    if _filter([comm]):
        text = comm.text
        for sect_idx, sect in enumerate(lun(comm.sectionList)):
            ts = sect.textSpan
            if ts is None:
                print(u"Section %s does not have a textSpan ")
                "field set" % (sect.uuid.uuidString)
                continue
            print(u"Section %d (%s)[kind: %s], from %d to %d:" % (
                sect_idx, sect.uuid.uuidString, sect.kind, ts.start, ts.ending))
            print(u"%s" % (text[ts.start:ts.ending]))
            print()
        print()


def print_situation_mentions(comm, tool=None, situation_mention_set_filter=None):
    """
    Print information for all :class:`.SituationMention`s (some of which
    may not have a :class:`.Situation`)

    Args:
        comm (Communication):
        tool (str): Deprecated.
                    If not `None`, only print information for
                    :class:`.SituationMention` objects with a matching
                    `metadata.tool` field
        situation_mention_set_filter (func): If not None, print information
            for only those :class:`.SituationMentionSet` objects that pass
            this filter.  Should be a function that takes a list of
            annotations (objects with metadata fields) and returns a
            list of annotations (possibly filtered and re-ordered).
    """
    _filter = filter_unnone(tool_to_filter(tool, situation_mention_set_filter))
    for sm_set_idx, sm_set in enumerate(lun(comm.situationMentionSetList)):
        if _filter([sm_set]):
            print(u"Situation Set %d (%s):" % (sm_set_idx,
                                               sm_set.metadata.tool))
            for sm_idx, sm in enumerate(sm_set.mentionList):
                print(u"  SituationMention %d-%d:" % (sm_set_idx, sm_idx))
                _print_situation_mention(sm)
                print()
            print()


def print_situations(comm, tool=None, situation_set_filter=None):
    """
    Print information for all :class:`.Situation` objects and their
    associated :class:`.SituationMention` objects

    Args:
        comm (Communication):
        tool (str): Deprecated.
                    If not `None`, only print information for
                    :class:`.Situation` objects with a matching
                    `metadata.tool` field
        situation_set_filter (func): If not None, print information
            for only those :class:`.SituationSet` objects that pass
            this filter.  Should be a function that takes a list of
            annotations (objects with metadata fields) and returns a
            list of annotations (possibly filtered and re-ordered).
    """
    _filter = filter_unnone(tool_to_filter(tool, situation_set_filter))
    for s_set_idx, s_set in enumerate(lun(comm.situationSetList)):
        if _filter([s_set]):
            print(u"Situation Set %d (%s):" % (s_set_idx,
                                               s_set.metadata.tool))
            for s_idx, situation in enumerate(s_set.situationList):
                print(u"  Situation %d-%d:" % (s_set_idx, s_idx))
                _p(6, 18, u"situationType", situation.situationType)
                for sm_idx, sm in enumerate(lun(situation.mentionList)):
                    print(u" " * 6 + u"SituationMention %d-%d-%d:" % (
                        s_set_idx, s_idx, sm_idx))
                    _print_situation_mention(sm)
                print()
            print()


def _print_situation_mention(situationMention):
    """
    Print SituationMention information needed to display both
    Situation and SituationMention types.

    Args:
        situationMention (SituationMention):
    """
    if situationMention.text:
        _p(10, 20, u"text", situationMention.text)
    if situationMention.situationType:
        _p(10, 20, u"situationType", situationMention.situationType)
    if situationMention.situationKind:
        _p(10, 20, u"situationKind", situationMention.situationKind)
    if situationMention.intensity:
        _p(10, 20, u"intensity", str(situationMention.intensity))
    for arg_idx, ma in enumerate(lun(situationMention.argumentList)):
        print(u" " * 10 + u"Argument %d:" % arg_idx)
        if ma.role:
            _p(14, 16, u"role", ma.role)
        if ma.entityMention:
            _p(14, 16, u"entityMention",
                u" ".join(_get_tokens_for_entityMention(ma.entityMention)))
        if ma.propertyList:
            # PROTO-ROLE PROPERTIES: Format a separate list for each
            # distinct annotator (metadata.tool) which tool should be
            # either None or a string. Sort by annotator
            # (metadata.tool) and then by property (p.value)
            last_tool = False
            for p in sorted(ma.propertyList,
                            key=lambda x: (x.metadata.tool, x.value)):
                tool = p.metadata.tool
                if tool != last_tool:
                    print(u" " * 14 + u"Properties (%s):" % tool)
                    last_tool = tool
                _p(18, 20, p.value, u"%1.1f" % p.polarity)
        # A SituationMention can have an argumentList with a
        # MentionArgument that points to another SituationMention---
        # which could conceivably lead to loops.  We currently don't
        # traverse the list recursively, instead looking at only
        # SituationMentions referenced by top-level SituationMentions
        if ma.situationMention:
            print(u" " * 14 + u"situationMention:")
            if situationMention.text:
                _p(18, 20, u"text", situationMention.text)
            if situationMention.situationType:
                _p(18, 20, u"situationType", situationMention.situationType)


def _p(indent_level, justified_width, fieldname, content):
    """
    Print field-value pair, indented and justified.

    Args:
        indent_level (int): number of spaces by which to prefix output
        justified_width (int): number of characters fieldname and colon
            should occupy (justified on left, padded with spaces)
        fieldname (str): field name
        content (str): field value
    """
    print(
        (u" " * indent_level) +
        (fieldname + u":").ljust(justified_width) +
        content
    )


def print_text_for_communication(comm, tool=None, communication_filter=None):
    """Print `text field of :class:`.Communication`

    Args:
        comm (Communication):
        tool (str): Deprecated.
                    If not `None`, only print `text` field of
                    :class:`.Communication` objects with a matching
                    `metadata.tool` field
        communication_filter (func): If not None, print information
            for only those :class:`.Communication` objects that pass
            this filter.  Should be a function that takes a list of
            annotations (objects with metadata fields) and returns a
            list of annotations (possibly filtered and re-ordered).
    """
    _filter = filter_unnone(tool_to_filter(tool, communication_filter))
    if _filter([comm]):
        print(comm.text)


def print_id_for_communication(comm, tool=None, communication_filter=None):
    """Print ID field of :class:`.Communication`

    Args:
        comm (Communication):
        tool (str): Deprecated.
                    If not `None`, only print ID of
                    :class:`.Communication` objects with a matching
                    `metadata.tool` field
        communication_filter (func): If not None, print information
            for only those :class:`.Communication` objects that pass
            this filter.  Should be a function that takes a list of
            annotations (objects with metadata fields) and returns a
            list of annotations (possibly filtered and re-ordered).

    """
    _filter = filter_unnone(tool_to_filter(tool, communication_filter))
    if _filter([comm]):
        print(comm.id)


def print_communication_taggings_for_communication(
        comm, tool=None, communication_tagging_filter=None):
    """Print information for :class:`.CommunicationTagging` objects

    Args:
        comm (Communication):
        tool (str): Deprecated.
                    If not `None`, only print information for
                    :class:`.CommunicationTagging` objects with a
                    matching `metadata.tool` field
        communication_tagging_filter (func): If not None, print information
            for only those :class:`.CommunicationTagging` objects that pass
            this filter.  Should be a function that takes a list of
            annotations (objects with metadata fields) and returns a
            list of annotations (possibly filtered and re-ordered).
    """
    _filter = filter_unnone(tool_to_filter(
        tool, communication_tagging_filter))
    for tagging in _filter(lun(comm.communicationTaggingList)):
        print('%s: %s' % (
            tagging.taggingType,
            ' '.join('%s:%.3f' % p for p in
                     zip(tagging.tagList, tagging.confidenceList))
        ))


def print_tokens_with_entityMentions(comm, tool=None, entity_mention_set_filter=None):
    """Print information for :class:`.Token` objects that are part of an :class:`.EntityMention`

    Args:
        comm (Communication):
        tool (str): Deprecated.
                    If not `None`, only print information for tokens
                    that are associated with an
                    :class:`.EntityMention` that is part of an
                    :class:`.EntityMentionSet` with a matching
                    `metadata.tool` field
        entity_mention_set_filter (func): If not None, print information
            for only those :class:`.EntityMentionSet` objects that pass
            this filter.  Should be a function that takes a list of
            annotations (objects with metadata fields) and returns a
            list of annotations (possibly filtered and re-ordered).
    """
    _filter = tool_to_filter(tool, entity_mention_set_filter)

    em_by_tkzn_id = _get_entityMentions_by_tokenizationId(
        comm, entity_mention_set_filter=_filter)
    em_entity_num = _get_entity_number_for_entityMention_uuid(
        comm, entity_mention_set_filter=_filter)
    tokenizations_by_section = _get_tokenizations_grouped_by_section(comm)

    for tokenizations_in_section in tokenizations_by_section:
        for tokenization in tokenizations_in_section:
            if tokenization.tokenList:
                text_tokens = [token.text
                               for token
                               in tokenization.tokenList.tokenList]
                u = tokenization.uuid.uuidString
                if u in em_by_tkzn_id:
                    for em in em_by_tkzn_id[u]:
                        first_token_index = em.tokens.tokenIndexList[0]
                        last_token_index = em.tokens.tokenIndexList[-1]
                        entity_number = em_entity_num[em.uuid.uuidString]
                        text_tokens[first_token_index] = (
                            u"<ENTITY ID=%d>%s" %
                            (entity_number, text_tokens[first_token_index])
                        )
                        text_tokens[last_token_index] = (
                            u"%s</ENTITY>" % text_tokens[last_token_index]
                        )
                print(u" ".join(text_tokens))
        print()


def print_tokens_for_communication(comm, tool=None, tokenization_filter=None):
    """Print token text for a :class:`.Communication`

    Args:
        comm (Communication):
        tool (str): Deprecated.
                    If not `None`, only print token text for
                    :class:`.Communication` objects with a matching
                    `metadata.tool` field
        tokenization_filter (func): If not None, print information
            for only those :class:`.Tokenization` objects that pass
            this filter.  Should be a function that takes a list of
            annotations (objects with metadata fields) and returns a
            list of annotations (possibly filtered and re-ordered).
    """
    tokenizations_by_section = _get_tokenizations_grouped_by_section(
        comm, tokenization_filter=tool_to_filter(tool, tokenization_filter))

    for tokenizations_in_section in tokenizations_by_section:
        for tokenization in tokenizations_in_section:
            if tokenization.tokenList:
                text_tokens = [token.text
                               for token
                               in tokenization.tokenList.tokenList]
                print(u" ".join(text_tokens))
        print()


def print_penn_treebank_for_communication(comm, tool=None, parse_filter=None):
    """Print Penn-Treebank parse trees for all :class:`.Tokenization` objects

    Args:
        comm (Communication):
        tool (str): Deprecated.
                    If not `None`, only print information for
                    :class:`.Tokenization` objects with a matching
                    `metadata.tool` field
        parse_filter (func): If not None, print information
            for only those :class:`.Parse` objects that pass
            this filter.  Should be a function that takes a list of
            annotations (objects with metadata fields) and returns a
            list of annotations (possibly filtered and re-ordered).
    """
    _filter = filter_unnone(tool_to_filter(tool, parse_filter))
    for tokenization in get_tokenizations(comm):
        for parse in _filter(lun(tokenization.parseList)):
            print(penn_treebank_for_parse(parse) + u"\n\n")


def penn_treebank_for_parse(parse):
    """
    Return a Penn-Treebank style representation of a Parse object

    Args:
        parse (Parse):

    Returns:
        str: A string containing a Penn Treebank style parse tree
        representation
    """
    def _traverse_parse(nodes, node_index, indent=0):
        s = u""
        indent += len(nodes[node_index].tag) + 2
        if nodes[node_index].childList:
            s += u"(%s " % nodes[node_index].tag
            for i, child_node_index in enumerate(nodes[node_index].childList):
                if i > 0:
                    s += u"\n" + u" " * indent
                s += _traverse_parse(nodes, child_node_index, indent)
            s += u")"
        else:
            s += nodes[node_index].tag
        return s

    sorted_nodes = sorted(parse.constituentList, key=attrgetter('id'))
    return _traverse_parse(sorted_nodes, 0)


def _get_char_offset_tags_for_tokenization(comm, tokenization):
    '''
    Return list of `comm.text` substrings corresponding to the tokens in
    :class:`.Tokenization` `tokenization` (where tokens without
    `textSpan` fields are represented by empty strings in the output list).

    Args:
        comm (Communication):
        tokenization (Tokenization):
    '''
    if tokenization.tokenList:
        char_offset_tags = [''] * len(tokenization.tokenList.tokenList)

        if comm.text:
            for i, token in enumerate(tokenization.tokenList.tokenList):
                if token.textSpan:
                    char_offset_tags[i] = comm.text[
                        token.textSpan.start:token.textSpan.ending]
        return char_offset_tags


def _get_ending_tags_for_tokenization(comm, tokenization):
    '''
    Return list of numbers (as strings) corresponding to the ending
    offsets of tokens in :class:`.Tokenization` `tokenization` in the
    communication text (where tokens without `textSpan` fields are
    represented by None in the output list), or return None if
    `tokenization` is None.

    Args:
        comm (Communication):
        tokenization (Tokenization):
    '''
    if tokenization.tokenList:
        ending_tags = [None] * len(tokenization.tokenList.tokenList)

        if comm.text:
            for i, token in enumerate(tokenization.tokenList.tokenList):
                if token.textSpan:
                    ending_tags[i] = str(token.textSpan.ending)
        return ending_tags


def _get_start_tags_for_tokenization(comm, tokenization):
    '''
    Return list of numbers (as strings) corresponding to the starting
    offsets of tokens in :class:`.Tokenization` `tokenization` in the
    communication text (where tokens without `textSpan` fields are
    represented by None in the output list), or return None if
    `tokenization` is None.

    Args:
        comm (Communication):
        tokenization (Tokenization):
    '''
    if tokenization.tokenList:
        start_tags = [None] * len(tokenization.tokenList.tokenList)

        if comm.text:
            for i, token in enumerate(tokenization.tokenList.tokenList):
                if token.textSpan:
                    start_tags[i] = str(token.textSpan.start)
        return start_tags


def _sorted_dep_lists_for_tokenization(tokenization,
                                       dependency_parse_filter=None):
    """
    Return list of lists of dependencies whose parses match the provided
    annotation filter.  Each list of dependencies has length
    is equal to the number of tokens in this tokenization's token list,
    where the element at index i is a dependency if there is a
    dependency whose dep field is i and None otherwise.

    Args:
        tokenization (Tokenization):
        dependency_parse_filter (func): If not None, ignore
            those :class:`.DependencyParse` objects that do not pass
            this filter.  Should be a function that takes a list of
            annotations (objects with metadata fields) and returns a
            list of annotations (possibly filtered and re-ordered).
    """
    _filter = filter_unnone(dependency_parse_filter)

    def _dep_lists():
        if tokenization.tokenList is not None:
            for dependency_parse in _filter(tokenization.dependencyParseList):
                sorted_dep_list = [None] * len(tokenization.tokenList.tokenList)
                for dep in dependency_parse.dependencyList:
                    sorted_dep_list[dep.dep] = dep
                yield sorted_dep_list

    return list(_dep_lists())


def _get_conll_tags_for_tokenization(tokenization, dependency_parse_filter=None):
    """
    Return a list of lists of CoNLL 'HEAD' and 'DEPREL' tag pairs for a
    tokenization

    In the CoNLL data format:

        http://ufal.mff.cuni.cz/conll2009-st/task-description.html

    the HEAD for a token is the (1-indexed) index of that token's
    parent token.  The root token of the dependency parse has a HEAD
    index of 0.

    Args:
        tokenization (Tokenization):
        dependency_parse_filter (func): If not None, ignore
            those :class:`.DependencyParse` objects that do not pass
            this filter.  Should be a function that takes a list of
            annotations (objects with metadata fields) and returns a
            list of annotations (possibly filtered and re-ordered).

    Returns:
        list: A list of lists of pairs, each of which contains a
              CoNLL HEAD tag and CoNLL DEPREL tag (as strings), one
              pair for each token in the supplied tokenization.  If a
              token does not have a HEAD tag (e.g. punctuation
              tokens), the HEAD tag is an empty string.  If a token
              does not have a DEPREL tag (e.g. punctuation tokens),
              the DEPREL tag is an empty string.
    """
    return [
        list(map(
            lambda dep: (
                unicode(
                    '' if dep is None else (
                        0 if dep.gov is None else dep.gov + 1
                    )
                ),
                unicode(
                    '' if dep is None else (
                        '' if dep.edgeType is None else dep.edgeType
                    )
                ),
            ),
            dep_list
        ))
        for dep_list in _sorted_dep_lists_for_tokenization(
            tokenization,
            dependency_parse_filter=dependency_parse_filter)
    ]


def _get_entityMentions_by_tokenizationId(comm, entity_mention_set_filter=None):
    """
    Return entity mentions for a Communication grouped by Tokenization
    UUID string

    Args:
        comm (Communication):
        entity_mention_set_filter (func): If not None, ignore
            those :class:`.EntityMentionSet` objects that do not pass
            this filter.  Should be a function that takes a list of
            annotations (objects with metadata fields) and returns a
            list of annotations (possibly filtered and re-ordered).

    Returns:
        A dictionary of lists of EntityMentions, where the dictionary
        keys are Tokenization UUID strings.
    """
    _filter = filter_unnone(entity_mention_set_filter)
    mentions_by_tkzn_id = defaultdict(list)
    for entitySet in lun(comm.entitySetList):
        for entity in entitySet.entityList:
            for entityMention in entity.mentionList:
                if _filter([entityMention.entityMentionSet]):
                    u = entityMention.tokens.tokenizationId.uuidString
                    mentions_by_tkzn_id[u].append(entityMention)
    return mentions_by_tkzn_id


def _get_entity_number_for_entityMention_uuid(comm, entity_mention_set_filter=None):
    """Create mapping from EntityMention UUID to (zero-indexed)
    'Entity Number'

    Args:
        comm (Communication):
        entity_mention_set_filter (func): If not None, ignore
            those :class:`.EntityMentionSet` objects that do not pass
            this filter.  Should be a function that takes a list of
            annotations (objects with metadata fields) and returns a
            list of annotations (possibly filtered and re-ordered).

    Returns:
        A dictionary where the keys are EntityMention UUID strings,
        and the values are "Entity Numbers", where the first Entity is
        assigned number 0, the second Entity is assigned number 1,
        etc.
    """
    _filter = filter_unnone(entity_mention_set_filter)
    entity_number_for_entityMention_uuid = {}
    entity_number_counter = 0

    if comm.entitySetList:
        for entitySet in comm.entitySetList:
            for entity in entitySet.entityList:
                any_mention = False
                for entityMention in entity.mentionList:
                    if _filter([entityMention.entityMentionSet]):
                        entity_number_for_entityMention_uuid[
                            entityMention.uuid.uuidString
                        ] = entity_number_counter
                    any_mention = True
                if any_mention:
                    entity_number_counter += 1
    return entity_number_for_entityMention_uuid


def _get_tokenizations_grouped_by_section(comm, tokenization_filter=None):
    """
    Return a list of lists of Tokenization objects in a Communication

    Args:
        comm (Communication):
        tokenization_filter (func): If not None, ignore
            those :class:`.Tokenization` objects that do not pass
            this filter.  Should be a function that takes a list of
            annotations (objects with metadata fields) and returns a
            list of annotations (possibly filtered and re-ordered).

    Returns:
        Returns a list of lists of Tokenization objects, where the
        Tokenization objects are grouped by Section
    """
    _filter = filter_unnone(tokenization_filter)

    return [
        _filter([
            sentence.tokenization
            for sentence in lun(section.sentenceList)
        ])
        for section in lun(comm.sectionList)
    ]


def _get_tokens_for_entityMention(entityMention):
    """
    Return list of token strings for an EntityMention

    Args:
        entityMention (EntityMention):

    Returns:
        A list of token strings
    """
    tokens = []
    for tokenIndex in entityMention.tokens.tokenIndexList:
        tokens.append(entityMention.tokens.tokenization.tokenList.tokenList[
                      tokenIndex].text)
    return tokens



from ..services.results import ResultsServerService
from .service_wrapper import (
    ConcreteServiceClientWrapper,
    ConcreteServiceWrapper,
    HTTPConcreteServiceClientWrapper,
    SubprocessConcreteServiceWrapper)


HTTPResultsServerClientWrapper = type(
    'HTTPResultsServerClientWrapper',
    (HTTPConcreteServiceClientWrapper,),
    {'concrete_service_class': ResultsServerService})


ResultsServerClientWrapper = type(
    'ResultsServerClientWrapper',
    (ConcreteServiceClientWrapper,),
    {'concrete_service_class': ResultsServerService})


ResultsServerServiceWrapper = type(
    'ResultsServerServiceWrapper',
    (ConcreteServiceWrapper,),
    {'concrete_service_class': ResultsServerService})


SubprocessResultsServerServiceWrapper = type(
    'SubprocessResultsServerServiceWrapper',
    (SubprocessConcreteServiceWrapper,),
    {'concrete_service_wrapper_class': ResultsServerServiceWrapper})



#!/usr/bin/env python

"""
Compare two Concrete files by converting to JSON then running the Git
diff command
"""
from __future__ import print_function
from __future__ import unicode_literals

import concrete.version
import argparse
import codecs
import os
import os.path
import subprocess
import tempfile
import logging

from concrete.util import communication_file_to_json
from concrete.util import set_stdout_encoding


def main():
    set_stdout_encoding()

    parser = argparse.ArgumentParser(
        description="Compare JSON representation of two concrete files")
    parser.add_argument('--include-uuids', action='store_true',
                        help="Include UUIDs in JSON output")
    parser.add_argument('--include-timestamps', action='store_true',
                        help="Include timestamps in JSON output")
    parser.add_argument('-l', '--loglevel', '--log-level',
                        help='Logging verbosity level threshold (to stderr)',
                        default='info')
    parser.add_argument('file_one')
    parser.add_argument('file_two')
    concrete.version.add_argparse_argument(parser)
    args = parser.parse_args()

    logging.basicConfig(format='%(asctime)-15s %(levelname)s: %(message)s',
                        level=args.loglevel.upper())

    tmp_path = tempfile.mkdtemp()

    json_one_filename = os.path.join(tmp_path, os.path.basename(args.file_one))
    json_two_filename = os.path.join(tmp_path, os.path.basename(args.file_two))

    # Prevent filename collision
    if json_one_filename == json_two_filename:
        json_two_filename += ".1"

    json_comm_one = communication_file_to_json(
        args.file_one, remove_timestamps=not args.include_timestamps,
        remove_uuids=not args.include_uuids)
    json_comm_two = communication_file_to_json(
        args.file_two, remove_timestamps=not args.include_timestamps,
        remove_uuids=not args.include_uuids)

    codecs.open(json_one_filename, "w", encoding="utf-8").write(json_comm_one)
    codecs.open(json_two_filename, "w", encoding="utf-8").write(json_comm_two)

    diff_command = "diff --unified %s %s" % (
        json_one_filename, json_two_filename)
    diff_output = subprocess.Popen(
        diff_command, shell=True, stdout=subprocess.PIPE).stdout.read()

    # Clean up temporary files
    os.remove(json_one_filename)
    os.remove(json_two_filename)
    os.rmdir(tmp_path)

    print(diff_output)


if __name__ == "__main__":
    main()
